==> erlang-ls-0.52.0.tar.gz: erlang_ls-0.52.0/apps/els_lsp/src/els_diagnostics.erl <==
%%==============================================================================
%% Behaviour and API to report diagnostics
%%==============================================================================
-module(els_diagnostics).

%%==============================================================================
%% Includes
%%==============================================================================
-include("els_lsp.hrl").
-include_lib("kernel/include/logger.hrl").

%%==============================================================================
%% Types
%%==============================================================================
-type diagnostic() :: #{
    range := range(),
    severity => severity(),
    code => number() | binary(),
    source => binary(),
    message := binary(),
    relatedInformation => [related_info()],
    data => binary()
}.
-type diagnostic_id() :: binary().
-type related_info() :: #{
    location := location(),
    message := binary()
}.
-type severity() ::
    ?DIAGNOSTIC_ERROR
    | ?DIAGNOSTIC_WARNING
    | ?DIAGNOSTIC_INFO
    | ?DIAGNOSTIC_HINT.
-export_type([
    diagnostic/0,
    diagnostic_id/0,
    severity/0
]).

%%==============================================================================
%% Callback Functions Definitions
%%==============================================================================
-callback is_default() -> boolean().
-callback run(uri()) -> [diagnostic()].
-callback source() -> binary().
-callback on_complete(uri(), [diagnostic()]) -> ok.
-optional_callbacks([on_complete/2]).

%%==============================================================================
%% API
%%==============================================================================
-export([
    available_diagnostics/0,
    default_diagnostics/0,
    enabled_diagnostics/0,
    make_diagnostic/4,
    make_diagnostic/5,
    run_diagnostics/1
]).

%%==============================================================================
%% API
%%==============================================================================

-spec available_diagnostics() -> [diagnostic_id()].
available_diagnostics() ->
    [
        <<"atom_typo">>,
        <<"bound_var_in_pattern">>,
        <<"compiler">>,
        <<"crossref">>,
        <<"dialyzer">>,
        <<"edoc">>,
        <<"gradualizer">>,
        <<"elvis">>,
        <<"unused_includes">>,
        <<"unused_macros">>,
        <<"unused_record_fields">>,
        <<"refactorerl">>,
        <<"eqwalizer">>
    ].

-spec default_diagnostics() -> [diagnostic_id()].
default_diagnostics() ->
    [Id || Id <- available_diagnostics(), (cb_module(Id)):is_default()].

-spec enabled_diagnostics() -> [diagnostic_id()].
enabled_diagnostics() ->
    Config = els_config:get(diagnostics),
    Default = default_diagnostics(),
    Enabled = maps:get("enabled", Config, []),
    Disabled = maps:get("disabled", Config, []),
    lists:usort((Default ++ valid(Enabled)) -- valid(Disabled)).

-spec make_diagnostic(range(), binary(), severity(), binary()) ->
    diagnostic().
make_diagnostic(Range, Message, Severity, Source) ->
    #{
        range => Range,
        message => Message,
        severity => Severity,
        source => Source
    }.

-spec make_diagnostic(range(), binary(), severity(), binary(), binary()) ->
    diagnostic().
make_diagnostic(Range, Message, Severity, Source, Data) ->
    #{
        range => Range,
        message => Message,
        severity => Severity,
        source => Source,
        data => Data
    }.

-spec run_diagnostics(uri()) -> [pid()].
run_diagnostics(Uri) ->
    [run_diagnostic(Uri, Id) || Id <- enabled_diagnostics()].

%%==============================================================================
%% Internal Functions
%%==============================================================================

-spec run_diagnostic(uri(), diagnostic_id()) -> pid().
run_diagnostic(Uri, Id) ->
    CbModule = cb_module(Id),
    Source = CbModule:source(),
    Module = atom_to_binary(els_uri:module(Uri), utf8),
    Title = <<Source/binary, " (", Module/binary, ")">>,
    Start = erlang:monotonic_time(millisecond),
    Config = #{
        task => fun(U, _) -> CbModule:run(U) end,
        entries => [Uri],
        title => Title,
        on_complete =>
            fun(Diagnostics) ->
                case erlang:function_exported(CbModule, on_complete, 2) of
                    true ->
                        CbModule:on_complete(Uri, Diagnostics);
                    false ->
                        ok
                end,
                els_diagnostics_provider:notify(Diagnostics, self()),
                End = erlang:monotonic_time(millisecond),
                Duration = End - Start,
                Event = #{
                    uri => Uri,
                    source => Source,
                    duration_ms => Duration,
                    number_of_diagnostics => length(Diagnostics),
                    type => <<"diagnostics">>
                },
                ?LOG_DEBUG("Diagnostics completed. [event=~p]", [Event]),
                els_telemetry:send_notification(Event)
            end
    },
    {ok, Pid} = els_background_job:new(Config),
    Pid.

%% @doc Return the callback module for a given Diagnostic Identifier
-spec cb_module(diagnostic_id()) -> module().
cb_module(Id) ->
    binary_to_existing_atom(<<"els_", Id/binary, "_diagnostics">>, utf8).

-spec is_valid(diagnostic_id()) -> boolean().
is_valid(Id) ->
    lists:member(Id, available_diagnostics()).

-spec valid([string()]) -> [diagnostic_id()].
valid(Ids0) ->
    Ids = [els_utils:to_binary(Id) || Id <- Ids0],
    {Valid, Invalid} = lists:partition(fun is_valid/1, Ids),
    case Invalid of
        [] ->
            ok;
        _ ->
            Fmt = "Skipping invalid diagnostics in config file: ~p",
            Args = [Invalid],
            Msg = lists:flatten(io_lib:format(Fmt, Args)),
            ?LOG_WARNING(Msg),
            els_server:send_notification(
                <<"window/showMessage">>,
                #{
                    type => ?MESSAGE_TYPE_WARNING,
                    message => els_utils:to_binary(Msg)
                }
            )
    end,
    Valid.

==> erlang-ls-1.1.0.tar.gz: erlang_ls-1.1.0/apps/els_lsp/src/els_diagnostics.erl <==
%%==============================================================================
%% Behaviour and API to report diagnostics
%%==============================================================================
-module(els_diagnostics).

%%==============================================================================
%% Includes
%%==============================================================================
-include("els_lsp.hrl").
-include_lib("kernel/include/logger.hrl").

%%==============================================================================
%% Types
%%==============================================================================
-type diagnostic() :: #{
    range := range(),
    severity => severity(),
    code => number() | binary(),
    source => binary(),
    message := binary(),
    relatedInformation => [related_info()],
    data => binary()
}.
-type diagnostic_id() :: binary().
-type related_info() :: #{
    location := location(),
    message := binary()
}.
-type severity() ::
    ?DIAGNOSTIC_ERROR
    | ?DIAGNOSTIC_WARNING
    | ?DIAGNOSTIC_INFO
    | ?DIAGNOSTIC_HINT.
-export_type([
    diagnostic/0,
    diagnostic_id/0,
    severity/0
]).

%%==============================================================================
%% Callback Functions Definitions
%%==============================================================================
-callback is_default() -> boolean().
-callback run(uri()) -> [diagnostic()].
-callback source() -> binary().
-callback on_complete(uri(), [diagnostic()]) -> ok.
-optional_callbacks([on_complete/2]).

%%==============================================================================
%% API
%%==============================================================================
-export([
    available_diagnostics/0,
    default_diagnostics/0,
    enabled_diagnostics/0,
    make_diagnostic/4,
    make_diagnostic/5,
    run_diagnostics/1
]).

%%==============================================================================
%% API
%%==============================================================================

-spec available_diagnostics() -> [diagnostic_id()].
available_diagnostics() ->
    [
        <<"atom_typo">>,
        <<"bound_var_in_pattern">>,
        <<"compiler">>,
        <<"crossref">>,
        <<"dialyzer">>,
        <<"edoc">>,
        <<"gradualizer">>,
        <<"elvis">>,
        <<"unused_includes">>,
        <<"unused_macros">>,
        <<"unused_record_fields">>,
        <<"refactorerl">>,
        <<"eqwalizer">>,
        <<"eunit">>
    ].

-spec default_diagnostics() -> [diagnostic_id()].
default_diagnostics() ->
    [Id || Id <- available_diagnostics(), (cb_module(Id)):is_default()].

-spec enabled_diagnostics() -> [diagnostic_id()].
enabled_diagnostics() ->
    Config = els_config:get(diagnostics),
    Default = default_diagnostics(),
    Enabled = maps:get("enabled", Config, []),
    Disabled = maps:get("disabled", Config, []),
    lists:usort((Default ++ valid(Enabled)) -- valid(Disabled)).

-spec make_diagnostic(range(), binary(), severity(), binary()) ->
    diagnostic().
make_diagnostic(Range, Message, Severity, Source) ->
    #{
        range => Range,
        message => Message,
        severity => Severity,
        source => Source
    }.

-spec make_diagnostic(range(), binary(), severity(), binary(), binary()) ->
    diagnostic().
make_diagnostic(Range, Message, Severity, Source, Data) ->
    #{
        range => Range,
        message => Message,
        severity => Severity,
        source => Source,
        data => Data
    }.

-spec run_diagnostics(uri()) -> [pid()].
run_diagnostics(Uri) ->
    case is_initial_indexing_done() of
        true ->
            ok = wait_for_indexing_job(Uri),
            [run_diagnostic(Uri, Id) || Id <- enabled_diagnostics()];
        false ->
            ?LOG_INFO(
                "Initial indexing is not done, skip running diagnostics for ~p",
                [els_uri:module(Uri)]
            ),
            []
    end.

%%==============================================================================
%% Internal Functions
%%==============================================================================
-spec is_initial_indexing_done() -> boolean().
is_initial_indexing_done() ->
    %% Keep in sync with els_indexing
    Jobs = [<<"Applications">>, <<"OTP">>, <<"Dependencies">>],
    JobTitles = els_background_job:list_titles(),
    lists:all(
        fun(Job) ->
            not lists:member(
                <<"Indexing ", Job/binary>>,
                JobTitles
            )
        end,
        Jobs
    ).

-spec wait_for_indexing_job(uri()) -> ok.
wait_for_indexing_job(Uri) ->
    %% Add delay to allowing indexing job to start
    timer:sleep(10),
    JobTitles = els_background_job:list_titles(),
    case lists:member(<<"Indexing ", Uri/binary>>, JobTitles) of
        false ->
            %% No indexing job is running, we're ready!
            ok;
        true ->
            %% Indexing job is still running, retry until it finishes
            wait_for_indexing_job(Uri)
    end.

-spec run_diagnostic(uri(), diagnostic_id()) -> pid().
run_diagnostic(Uri, Id) ->
    CbModule = cb_module(Id),
    Source = CbModule:source(),
    Module = atom_to_binary(els_uri:module(Uri), utf8),
    Title = <<Source/binary, " (", Module/binary, ")">>,
    Start = erlang:monotonic_time(millisecond),
    Config = #{
        task => fun(U, _) -> CbModule:run(U) end,
        entries => [Uri],
        title => Title,
        on_complete =>
            fun(Diagnostics) ->
                case erlang:function_exported(CbModule, on_complete, 2) of
                    true ->
                        CbModule:on_complete(Uri, Diagnostics);
                    false ->
                        ok
                end,
                els_diagnostics_provider:notify(Diagnostics, self()),
                End = erlang:monotonic_time(millisecond),
                Duration = End - Start,
                Event = #{
                    uri => Uri,
                    source => Source,
                    duration_ms => Duration,
                    number_of_diagnostics => length(Diagnostics),
                    type => <<"diagnostics">>
                },
                ?LOG_DEBUG("Diagnostics completed. [event=~p]", [Event]),
                els_telemetry:send_notification(Event)
            end
    },
    {ok, Pid} = els_background_job:new(Config),
    Pid.

%% @doc Return the callback module for a given Diagnostic Identifier
-spec cb_module(diagnostic_id()) -> module().
cb_module(Id) ->
    binary_to_existing_atom(<<"els_", Id/binary, "_diagnostics">>, utf8).

-spec is_valid(diagnostic_id()) -> boolean().
is_valid(Id) ->
    lists:member(Id, available_diagnostics()).

-spec valid([string()]) -> [diagnostic_id()].
valid(Ids0) ->
    Ids = [els_utils:to_binary(Id) || Id <- Ids0],
    {Valid, Invalid} = lists:partition(fun is_valid/1, Ids),
    case Invalid of
        [] ->
            ok;
        _ ->
            Fmt = "Skipping invalid diagnostics in config file: ~p",
            Args = [Invalid],
            Msg = lists:flatten(io_lib:format(Fmt, Args)),
            ?LOG_WARNING(Msg),
            els_server:send_notification(
                <<"window/showMessage">>,
                #{
                    type => ?MESSAGE_TYPE_WARNING,
                    message => els_utils:to_binary(Msg)
                }
            )
    end,
    Valid.
